#!/usr/bin/perl -w

use strict;
use Getopt::Long "GetOptions";
use FindBin qw($Bin);
use File::Spec::Functions;
use File::Basename;

$ENV{"LC_ALL"} = "C";

my $_FIRST_STEP;
my $_LAST_STEP; 

my $_TRAIN_SRC;
my $_TRAIN_TRG;
my $_TRAIN_SRC_RULES;
my $_TRAIN_SRC_BM;
my $_TRAIN_TRG_BM;
my $_TRAIN_SRC_UNF;
my $_ALIGNMENT_FILE;

my $_OUTPUT_DIR;

my $_MAX_TUPLE_LENGTH="5"; 
my $_MAX_TUPLE_FERT="4"; 
my $_TUPLE_NBEST="30"; 
my $_MAX_RULE_COST="4"; 
my $_MAX_RULE_LENGTH="10"; 
my $_OPTIONS_BM="-order_3_-unk_-gt3min_1_-kndiscount_-interpolate";
my $_SMOOTH="0";
my $_RECURSIVE="";
my $_NOCOSTMON="";
my $_NAME_SRC_BM="";
my $_NAME_TRG_BM="";
my $_OPTIONS_SM="-order_3_-unk_-gt3min_1_-kndiscount_-interpolate";
my $_NAME_SRC_UNF="";

my $_REPLACE = 0;
my $_HELP = 0;

my $_BINCODER = $ENV{'BINCODER'};
my $_SRIBIN = $ENV{'SRILM'};
my $_KENBIN = $ENV{'KENLM'};
my $_BIN="$_BINCODER/bin";

my $cmd;
my $usage="USAGE: training.perl 
   --first-step       i 
   --last-step        i
   --output-dir       s : output directory

   --train-src        s : training (bitext) source (raw) file
   --train-trg        s : training (bitext) target (raw) file
   --train-src-rules  s : training (bitext) source file for rules (tagged)
   --train-src-bm     s : training (bitext) source file for bm (factors used)
   --train-trg-bm     s : training (bitext) target file for bm (factors used)
   --train-src-unf    s : training (bitext) source file for unfolded srclm
   --alignment-file   s : alignment file

   --max-tuple-length i : max tuple length in words (source or target)               [def $_MAX_TUPLE_LENGTH]
   --max-tuple-fert   f : max tuple fertility (source Vs target words)               [def $_MAX_TUPLE_FERT]
   --tuple-nbest      i : n-best translation options are kept (p(u|v)*p(v|u))        [def $_TUPLE_NBEST]
   --max-rule-length  f : max lenght of rules                                        [def $_MAX_RULE_LENGTH]

   --options-bm       s : string with language model options                         [def $_OPTIONS_BM]
   --options-sm       s : string with ufolded source language model options          [def $_OPTIONS_SM]
   --name-src-bm      s : description tag for bilingual language model
   --name-trg-bm      s : description tag for bilingual language model
   --name-src-unf     s : description tag for monolingual src language model

   --bin              s : to override bincoder directory                             [def $_BIN]
   --sribin           s : srilm directory                                            [def $_SRIBIN]
   --kenbin           s : kenlm directory                                            [def $_KENBIN]
   --replace            : replace existing files                                     [def $_REPLACE]
   --help               : this help

Steps:
(0) estimate word-to-word probabilities
(1) extract tuples
(2) solve source-NULLed tuples
(3) prune tuples & estimate frequency/lexical scores
(4) estimate bilingual LM
(5) estimate reordering rules
(6) estimate lexicalized reordering
(7) estimate unfolded source lm
(8) clean files
";

$_HELP = 1 unless &GetOptions('first-step=i' => \$_FIRST_STEP,
		       'last-step=i' => \$_LAST_STEP,
		       'train-src=s' => \$_TRAIN_SRC,
		       'train-trg=s' => \$_TRAIN_TRG,
		       'train-src-rules=s' => \$_TRAIN_SRC_RULES,
		       'train-src-bm=s' => \$_TRAIN_SRC_BM,
		       'train-trg-bm=s' => \$_TRAIN_TRG_BM,
		       'train-src-unf=s' => \$_TRAIN_SRC_UNF,
		       'name-src-bm=s' => \$_NAME_SRC_BM,
		       'name-trg-bm=s' => \$_NAME_TRG_BM,
		       'name-src-unf=s' => \$_NAME_SRC_UNF,
		       'alignment-file=s' => \$_ALIGNMENT_FILE,
		       'output-dir=s' => \$_OUTPUT_DIR,
		       'max-tuple-length=i' => \$_MAX_TUPLE_LENGTH,
		       'max-tuple-fert=f' => \$_MAX_TUPLE_FERT,
		       'tuple-nbest=i' => \$_TUPLE_NBEST,
		       'max-rule-length=f' => \$_MAX_RULE_LENGTH,
		       'options-bm=s' => \$_OPTIONS_BM,
		       'options-sm=s' => \$_OPTIONS_SM,
		       'bin=s' => \$_BIN,
		       'sribin=s' => \$_SRIBIN,
		       'kenbin=s' => \$_KENBIN,
		       'replace' => \$_REPLACE,
		       'help' => \$_HELP);

print STDERR "   ==========================================
     training.perl: training bincode models
     jmcrego[at]limsi[dot]fr, June 2011        
   ==========================================
";
  
die $usage if ($_HELP);
die "ERROR: undefined --first-step or --last-step parameters\n".$usage unless (defined $_FIRST_STEP && defined $_LAST_STEP);
die "ERROR: undefined --output-dir parameter\n".$usage unless (defined $_OUTPUT_DIR);

my $TAGBM=$_OPTIONS_BM; 
$_OPTIONS_BM=~s/\_/ /g;
my $TAGSM=$_OPTIONS_SM; 
$_OPTIONS_SM=~s/\_/ /g;

### output files
my $_LEX      ="$_OUTPUT_DIR/lex";
my $_LEX1     ="$_OUTPUT_DIR/lex.f2n";
my $_LEX2     ="$_OUTPUT_DIR/lex.n2f";
my $_UNITS    ="$_OUTPUT_DIR/unfold";
my $_UNITSNULL="$_OUTPUT_DIR/unfoldNULL";
my $_RULES    ="$_OUTPUT_DIR/posrules.max$_MAX_RULE_LENGTH.smooth$_SMOOTH.$_RECURSIVE.$_NOCOSTMON";
my $_VOC1     ="$_OUTPUT_DIR/unfold.maxs$_MAX_TUPLE_LENGTH.maxf$_MAX_TUPLE_FERT.voc";
my $_VOC1_r1r2="$_OUTPUT_DIR/unfold.maxs$_MAX_TUPLE_LENGTH.maxf$_MAX_TUPLE_FERT.voc.rfreq1.rfreq2";
my $_VOC2     ="$_OUTPUT_DIR/unfold.maxs$_MAX_TUPLE_LENGTH.maxf$_MAX_TUPLE_FERT.tnb$_TUPLE_NBEST.voc";
my $_VOC2_l1l2="$_OUTPUT_DIR/unfold.maxs$_MAX_TUPLE_LENGTH.maxf$_MAX_TUPLE_FERT.tnb$_TUPLE_NBEST.voc.lex1.lex2";
my $_VOC2_r1r2="$_OUTPUT_DIR/unfold.maxs$_MAX_TUPLE_LENGTH.maxf$_MAX_TUPLE_FERT.tnb$_TUPLE_NBEST.voc.rfreq1.rfreq2";
my $_VOC2_l1l2r1r2="$_OUTPUT_DIR/unfold.maxs$_MAX_TUPLE_LENGTH.maxf$_MAX_TUPLE_FERT.tnb$_TUPLE_NBEST.voc.lex1.lex2.rfreq1.rfreq2";
my $_LEXRM    ="$_OUTPUT_DIR/unfold.maxs$_MAX_TUPLE_LENGTH.maxf$_MAX_TUPLE_FERT.tnb$_TUPLE_NBEST.voc.msdcfb";
my $_BILLM    ="$_OUTPUT_DIR/unfold.maxs$_MAX_TUPLE_LENGTH.maxf$_MAX_TUPLE_FERT.tnb$_TUPLE_NBEST.voc.${_NAME_SRC_BM}-${_NAME_TRG_BM}.bil.$TAGBM.lm";
my $_UNFTRN   ="$_OUTPUT_DIR/unfoldNULL.$_NAME_SRC_UNF";
my $_SRCLM    ="$_OUTPUT_DIR/unfoldNULL.$_NAME_SRC_UNF.$TAGSM.lm";

if (!-d $_OUTPUT_DIR) {
     $cmd="mkdir -p $_OUTPUT_DIR";
     safesystem("$cmd") or die "ERROR: $cmd failed!\n";
}

### MAIN

print STDERR "training starts @ ".`date`;
&lexicon_model()                           if $_FIRST_STEP==0;
&extract_tuples()                          if $_FIRST_STEP<=1 && $_LAST_STEP>=1;
&solve_src_nulls()                         if $_FIRST_STEP<=2 && $_LAST_STEP>=2;
&prune_tuples_estimate_freq_estimate_lex() if $_FIRST_STEP<=3 && $_LAST_STEP>=3;
&estimate_bilingual_lm()                   if $_FIRST_STEP<=4 && $_LAST_STEP>=4;
&estimate_reordering_rules()               if $_FIRST_STEP<=5 && $_LAST_STEP>=5;
&estimate_lexicalized_reordering()         if $_FIRST_STEP<=6 && $_LAST_STEP>=6;
&estimate_unfolded_lm()                    if $_FIRST_STEP<=7 && $_LAST_STEP>=7;
&clean_files()                             if $_FIRST_STEP<=8 && $_LAST_STEP>=8;
print STDERR "training finished @ ".`date`;
exit;

############################################################################################################
### (0) lexicon model
############################################################################################################

sub lexicon_model {
    print STDERR "\n(0) lexicon model @ ".`date`;
    die("ERROR: Cannot find alignment-file: $_ALIGNMENT_FILE !!\n") unless (-r $_ALIGNMENT_FILE);
    die("ERROR: Cannot find training source file: $_TRAIN_SRC !!\n") unless (-r $_TRAIN_SRC);
    die("ERROR: Cannot find training target file: $_TRAIN_TRG !!\n") unless (-r $_TRAIN_TRG);

    if ($_REPLACE || !-e $_LEX1 || !-e $_LEX2){
	$cmd = "$_BIN/alig2lex.perl -src $_TRAIN_SRC -trg $_TRAIN_TRG -ali $_ALIGNMENT_FILE -lex $_LEX";
	safesystem("$cmd") or die "ERROR: lexicon model failed!\n";
    }
}
 
############################################################################################################
### (1) extract tuples
############################################################################################################

sub extract_tuples {
    print STDERR "\n(1) extract tuples @ ".`date`;
    die("ERROR: Cannot find alignment-file: $_ALIGNMENT_FILE !!\n") unless (-r $_ALIGNMENT_FILE);
    die("ERROR: Cannot find training source file: $_TRAIN_SRC !!\n") unless (-r $_TRAIN_SRC);
    die("ERROR: Cannot find training target file: $_TRAIN_TRG !!\n") unless (-r $_TRAIN_TRG);
    die("ERROR: Cannot find training source tag file: $_TRAIN_SRC_RULES !!\n") unless (-r $_TRAIN_SRC_RULES);

    if ($_REPLACE || !-e ${_UNITS}."NULL"){
	$cmd = "$_BIN/alig2unf -ali $_ALIGNMENT_FILE -swrd $_TRAIN_SRC -twrd $_TRAIN_TRG -stag $_TRAIN_SRC_RULES -keep > ${_UNITS}NULL";
	safesystem("$cmd") or die "ERROR: Tuple extraction failed!\n";
    }
}

############################################################################################################
### (2) solve source-NULLed tuples
############################################################################################################

sub solve_src_nulls {
    print STDERR "\n(2) solve src nulls @ ".`date`;
    die("ERROR: Cannot find units file: $_UNITSNULL !!\n") unless (-r ${_UNITSNULL});
    die("ERROR: Cannot find lex.f2e file: $_LEX1 !!\n") unless (-r $_LEX1);
    die("ERROR: Cannot find lex.e2f file: $_LEX2 !!\n") unless (-r $_LEX2);

    if ($_REPLACE || !-e ${_UNITS}){
	$cmd="$_BIN/srcNULLs -fibm11 $_LEX1 -fibm12 $_LEX2 -ftuples ${_UNITS}NULL > $_UNITS";
	safesystem("$cmd") or die "ERROR: srcNULLs failed!\n";
    }
}

############################################################################################################
### (3) prune tuples & estimate frequency/lexical scores
############################################################################################################

sub prune_tuples_estimate_freq_estimate_lex {
    print STDERR "\n(3) prune tuples \& estimate freq model @ ".`date`;
    die("ERROR: Cannot find units file: ${_UNITS} !!\n") unless (-r ${_UNITS});
    die("ERROR: Cannot find lex.f2e file: $_LEX1 !!\n") unless (-r $_LEX1);
    die("ERROR: Cannot find lex.e2f file: $_LEX2 !!\n") unless (-r $_LEX2);

    if ($_REPLACE || !-e ${_VOC1}){
	$cmd="$_BIN/prunetpl.perl -max_size $_MAX_TUPLE_LENGTH -max_fert $_MAX_TUPLE_FERT -voc $_VOC1 < $_UNITS";
	safesystem("$cmd") or die "ERROR: prunetpl failed!\n";
    }

    if ($_REPLACE || !-e ${_VOC2} || !-e $_VOC2_r1r2){
	$cmd="$_BIN/prunevoc.perl -tnb $_TUPLE_NBEST -voc $_VOC1 -scores $_VOC1_r1r2 -ovoc $_VOC2 -oscores $_VOC2_r1r2";
	safesystem("$cmd") or die "ERROR: prunevoc failed!\n";
    }

    if ($_REPLACE || !-e $_VOC2_l1l2){
	$cmd="$_BIN/feat_lex -table $_VOC2 -i_st $_LEX2 -i_ts $_LEX1 -o $_VOC2_l1l2";
	safesystem("$cmd") or die "ERROR: feat_lex failed!\n";
    }

    if ($_REPLACE || !-e $_VOC2_l1l2r1r2){
	$cmd="paste -d ' ' $_VOC2_l1l2 $_VOC2_r1r2 > $_VOC2_l1l2r1r2";
	safesystem("$cmd") or die "ERROR: paste failed!\n";
    }
}

############################################################################################################
### (4) estimate bilingual LM
############################################################################################################

sub estimate_bilingual_lm {
    print STDERR "\n(4) estimate bilingual LM @ ".`date`;
    die("ERROR: Cannot find units file: ${_UNITS} !!\n") unless (-r $_UNITS);
    die("ERROR: Cannot find units file: ${_UNITSNULL} !!\n") unless (-r $_UNITSNULL);
    die("ERROR: Cannot find voc2 file: $_VOC2 !!\n") unless (-r $_VOC2);
    die("ERROR: Cannot find training source tag file: $_TRAIN_SRC_BM !!\n") unless (-r $_TRAIN_SRC_BM);
    die("ERROR: Cannot find training target tag file: $_TRAIN_TRG_BM !!\n") unless (-r $_TRAIN_TRG_BM);
    die("ERROR: missing --name-src-bm parameter !!\n") unless ($_NAME_SRC_BM!~/^$/);
    die("ERROR: missing --name-trg-bm parameter !!\n") unless ($_NAME_TRG_BM!~/^$/);

    my $_NAMEBM="${_NAME_SRC_BM}-${_NAME_TRG_BM}";
    if ($_REPLACE || !-e $_VOC2.".${_NAME_SRC_BM}-${_NAME_TRG_BM}.bil.seq"){
	$cmd="$_BIN/unf+null.perl ${_UNITS} ${_UNITS}NULL | $_BIN/unf2fact -stag $_TRAIN_SRC_BM -ttag $_TRAIN_TRG_BM -voc $_VOC2 -name $_NAMEBM";
	safesystem("$cmd") or die "ERROR: unfold+null or/and unf2fact.perl failed!\n";
    }
    if ($_REPLACE || !-e $_BILLM){
	$cmd="$_SRIBIN/ngram-count -text $_VOC2.$_NAMEBM.bil.seq -vocab $_VOC2.$_NAMEBM.bil.seq.voc -lm $_BILLM $_OPTIONS_BM";
	safesystem("$cmd") or die "ERROR: ngram-count failed!\n";
    }
    if ($_REPLACE || !-e $_BILLM.".mmap"){
	$cmd="$_KENBIN/build_binary trie $_BILLM $_BILLM.mmap";
	safesystem("$cmd") or die "ERROR: build_binary failed!\n";
    }
}

############################################################################################################
### (5) estimate reordering rules
############################################################################################################

sub estimate_reordering_rules {
    print STDERR "\n(5) estimate reordering rules @ ".`date`;
    die("ERROR: Cannot find units file: $_UNITSNULL !!\n") unless (-r $_UNITSNULL);
    die("ERROR: Cannot find training source tag file: $_TRAIN_SRC_RULES !!\n") unless (-r $_TRAIN_SRC_RULES);

    if ($_REPLACE || !-e $_RULES){
	$cmd="$_BIN/posrules -trn $_TRAIN_SRC_RULES -unf $_UNITSNULL -max $_MAX_RULE_LENGTH -smooth $_SMOOTH $_RECURSIVE $_NOCOSTMON > $_RULES";
	safesystem("$cmd") or die "ERROR: posrules failed!\n";
    }
}

############################################################################################################
### (6) estimate lexicalized reordering
############################################################################################################

sub estimate_lexicalized_reordering {
    print STDERR "\n(6) estimate lexicalized reordering @ ".`date`;
    die("ERROR: Cannot find units file: ${_UNITS} !!\n") unless (-r $_UNITS);
    die("ERROR: Cannot find units file: ${_UNITSNULL} !!\n") unless (-r $_UNITSNULL);
    die("ERROR: Cannot find units file: ${_VOC2} !!\n") unless (-r $_VOC2);

    if ($_REPLACE || !-e $_LEXRM){
	$cmd="$_BIN/unf+null.perl ${_UNITS} ${_UNITS}NULL | $_BIN/lexreord.perl -voc $_VOC2 > $_LEXRM";
	safesystem("$cmd") or die "ERROR: unf+null.perl or/and lexreord.perl failed!\n";
    }	
}

############################################################################################################
### (7) estimate unfolded lm
############################################################################################################

sub estimate_unfolded_lm {
    print STDERR "\n(7) estimate unfolded lm @ ".`date`;
    die("ERROR: Cannot find trainig file: ${_TRAIN_SRC_UNF} !!\n") unless (-r $_TRAIN_SRC_UNF);
    die("ERROR: Cannot find units file: ${_UNITS}NULL !!\n") unless (-r $_UNITS."NULL");
    die("ERROR: missing --name-src-unf parameter !!\n") unless ($_NAME_SRC_UNF!~/^$/);

    if ($_REPLACE || !-e $_UNFTRN){
	$cmd="$_BIN/unf2utrn.perl -trn ${_TRAIN_SRC_UNF} < ${_UNITS}NULL > ${_UNFTRN}";
	safesystem("$cmd") or die "ERROR: unf2utrn.perl failed!\n";
    }	
    if ($_REPLACE || !-e $_SRCLM){
	$cmd="$_SRIBIN/ngram-count -text $_UNFTRN -lm $_SRCLM $_OPTIONS_SM";
	safesystem("$cmd") or die "ERROR: ngram-count failed!\n";
    }
    if ($_REPLACE || !-e $_SRCLM.".mmap"){
	$cmd="$_KENBIN/build_binary trie $_SRCLM $_SRCLM.mmap";
	safesystem("$cmd") or die "ERROR: build_binary failed!\n";
    }
}

############################################################################################################
### (8) clean files
############################################################################################################

sub clean_files {
    print STDERR "\n(8) clean files @ ".`date`;
    pack_file($_LEX1);
    pack_file($_LEX2);
    pack_file($_VOC1);
    pack_file($_VOC1_r1r2);
    pack_file($_VOC2_l1l2);
    pack_file($_VOC2_r1r2);
    pack_file($_UNITS);
    pack_file($_UNITSNULL);

    my (@FILE) = glob "$_VOC2.*.seq";
    my $file;
    foreach $file (@FILE){
	pack_file($file);
	pack_file($file.".voc");
    }
    (@FILE) = glob "$_VOC2.*.lm";
    foreach $file (@FILE){
	pack_file($file);
    }
}

############################################################################################################
############################################################################################################
############################################################################################################

sub pack_file{
    my $file = $_[0];
    if (-e $file){
	safesystem("gzip $file") or die "ERROR: gzip $file failed!\n";
    }
}

sub safesystem {
  print STDERR "Executing: @_\n";
  system(@_);
  if ($? == -1) {
      print STDERR "ERROR: Failed to execute: @_\n  $!\n";
      exit(1);
  }
  elsif ($? & 127) {
      printf STDERR "ERROR: Execution of: @_\n  died with signal %d, %s coredump\n",
          ($? & 127),  ($? & 128) ? 'with' : 'without';
      exit(1);
  }
  else {
    my $exitcode = $? >> 8;
    print STDERR "Exit code: $exitcode\n" if $exitcode;
    return ! $exitcode;
  }
}

